import unittest
import trsfile
import os
import binascii
import numpy
from pathlib import Path
from trsfile import TrsFile, Header, SampleCoding

class TestTrsFile(unittest.TestCase):
	def setUp(self):
		self.trs_file = trsfile.open(Path(__file__).parent / 'data' / '90x500xfloat.trs')

	def tearDown(self):
		self.trs_file.close()

	def test_length(self):
		"""Check if length is indeed equal to 90"""
		self.assertEqual(len(self.trs_file), 90)

	def test_sample_length(self):
		"""Check if sample length is indeed equal to 500"""
		self.assertEqual(len(self.trs_file[0]), 500)

	def test_sample_coding(self):
		"""Check if sample coding is indeed FLOAT"""
		self.assertEqual(self.trs_file[0].sample_coding, SampleCoding.FLOAT)

	def test_iterator(self):
		"""Check if we can iterate over the traces"""
		i = 0
		for trace in self.trs_file:
			i += 1
		self.assertEqual(i, len(self.trs_file))

	def test_read_only(self):
		"""Check if the TrsFile is read only"""
		with self.assertRaises(TypeError) as cm:
			self.trs_file[0] = None

		with self.assertRaises(TypeError) as cm:
			del self.trs_file[0]

	def test_slice(self):
		"""Check if we can slice the TrsFile file"""
		steps = 3
		traces = self.trs_file[::steps]
		self.assertEqual(len(traces), len(self.trs_file) // steps)

	def test_reverse(self):
		"""Check if reverse works"""
		for i, trace in zip(range(1, len(self.trs_file) + 1), self.trs_file.reverse()):
			self.assertTrue(numpy.array_equal(self.trs_file[len(self.trs_file) - i].samples, trace.samples))

if __name__ == '__main__':
	unittest.main()



import os
import mmap
import struct
import numpy

from .trace import Trace
from .common import Header, SampleCoding

class TrsFile:
	handle = None
	file_handle = None
	temp_folder = None
	headers = {}

	# All our magic function to support easy usage of the Trs file format
	def __init__(self, path):
		self.path = path

		self.file_handle = open(path, 'rb', 0) # Disable buffering
		self.handle = mmap.mmap(self.file_handle.fileno(), 0, access = mmap.ACCESS_READ)

		self.__initialize_headers()

	def __del__(self):
		# Sanity close, no harm from calling twice, but we do expect user calling close!
		self.close()

	def __iter__(self):
		""" reset pointer """
		self.iterator_index = -1
		return self

	def __next__(self):
		self.iterator_index = self.iterator_index + 1

		if self.iterator_index >= len(self):
			raise StopIteration
		return self[self.iterator_index]

	def __enter__(self):
		"""Called when entering a `with` block"""
		if self.handle is None or self.handle.closed:
			raise ValueError('I/O operation on closed file')
		return self

	def __exit__(self, *args):
		"""Called when exiting a `with` block"""
		self.close()

	def __repr__(self):
		return repr([self[i] for i in range(0, len(self))])

	def __len__(self):
		"""Returns the total number of traces"""
		return self.headers[Header.NUMBER_TRACES] if Header.NUMBER_TRACES in self.headers else 0

	def __delitem__(self, index):
		raise TypeError('cannot modify existing trace set')

	def __setitem__(self, index, trace):
		raise TypeError('cannot modify existing trace set')

	def __getitem__(self, index):
		# check for slicing
		if isinstance(index, slice):
			# No bounds checking when using slices, that's how python rolls!
			r = range(*index.indices(self.headers[Header.NUMBER_TRACES]))
		else:
			# Wrap around for negative indexes
			if index < 0:
				index = index % self.headers[Header.NUMBER_TRACES]

			# Check if we are still in bounds
			if index >= self.headers[Header.NUMBER_TRACES]:
				raise IndexError('list index out of range')

			r = range(index, index + 1)

		# Now read in all traces
		traces = []
		for i in r:
			# Seek to the beginning of the trace
			self.handle.seek(self.data_offset + i * self.trace_length)

			# Read the title
			if Header.TITLE_SPACE in self.headers:
				title = self.handle.read(self.headers[Header.TITLE_SPACE]).decode('utf-8')
			else:
				title = Header.TRACE_TITLE.default

			# Read data
			if Header.LENGTH_DATA in self.headers:
				data = self.handle.read(self.headers[Header.LENGTH_DATA])
			else:
				data = bytes(Header.LENGTH_DATA.default)

			# Read all the samples
			samples = numpy.frombuffer(self.handle.read(self.trace_length), self.headers[Header.SAMPLE_CODING].format, self.headers[Header.NUMBER_SAMPLES])

			traces.append(Trace(self.headers[Header.SAMPLE_CODING], samples, data, title, self.headers))

		# Return the select item(s)
		if isinstance(index, slice):
			return traces
		else:
			# Earlier logic should ensure traces contains one element!
			return traces[0]

	def save(self):
		raise TypeError('cannot save existing trace set')

	def append(self, trace):
		raise TypeError('cannot modify existing trace set')

	def extend(self, traces):
		raise TypeError('cannot modify existing trace set')

	def insert(self, index, trace):
		raise TypeError('cannot modify existing trace set')

	def reverse(self):
		return self[::-1]

	def close(self):
		"""Closes the open file handle if it is opened"""

		if self.file_handle is not None and not self.file_handle.closed:
			self.file_handle.close()
		if self.handle is not None and not self.handle.closed:
			self.handle.close()

	def __initialize_headers(self):
		"""Read all internal headers from the file"""
		self.headers = {}

		# Add default headers and values if new
		self.handle.seek(0)

		# Parse all headers until the TRACE_BLOCK
		while Header.TRACE_BLOCK not in self.headers:
			# Obtain the Tag
			tag = self.handle.read(1)[0]

			# Obtain the Length
			tag_length = self.handle.read(1)[0]

			if (tag_length & 0x80) != 0:
				tag_length = int.from_bytes(self.handle.read(tag_length & 0x7F), 'little')

			# Obtain the Value
			tag_value = self.handle.read(tag_length) if tag_length > 0 else None

			# Interpreter it
			header = None
			if Header.has_value(tag):
				header = Header(tag)
				if header.type is int:
					tag_value = int.from_bytes(tag_value, 'little')
				elif header.type is float:
					tag_value, = struct.unpack('<f', tag_value)
				elif header.type is bool:
					tag_value, = struct.unpack('<?', tag_value)
				elif header.type is str:
					tag_value = tag_value.decode('utf-8')
				elif header.type is SampleCoding:
					tag_value = SampleCoding(tag_value[0])
			else:
				print('Warning: tag {tag:02X} is not supported by the library, please let us know.'.format(tag=tag))

			self.headers[tag if header is None else header] = tag_value

		# Pre-compute some static information based on headers
		self.data_offset = self.handle.tell()
		self.sample_length = self.headers[Header.NUMBER_SAMPLES] * self.headers[Header.SAMPLE_CODING].size
		self.trace_length = self.sample_length + self.headers.get(Header.LENGTH_DATA, 0) + self.headers.get(Header.TITLE_SPACE, 0)

		# Sanity: Check if we have all mandatory headers, if not, throw an error if we are in reading mode!
		if not Header.get_mandatory().issubset(self.headers):
			raise IOError('trace set does not contain all mandatory headers')

		# Sanity: Check if the file has the proper size
		self.handle.seek(0, os.SEEK_END)
		file_size = self.handle.tell()
		if file_size != self.data_offset + self.headers[Header.NUMBER_TRACES] * self.trace_length:
			raise IOError('trace set has an unexpected length')



import binascii

from .common import Header, SampleCoding

class Trace:
	"""
	Single trace
	"""

	def __init__(self, sample_coding : SampleCoding, samples, data = b'', title = 'trace', headers = {}):
		self.title = title
		self.data = data if data is not None else b''
		self.samples = samples
		self.sample_coding = sample_coding
		self.headers = headers

	def __len__(self):
		"""Returns the total number of samples in this trace"""
		return len(self.samples)

	def __iter__(self):
		""" reset pointer """
		self.iterator_index = -1
		return self

	def __next__(self):
		self.iterator_index = self.iterator_index + 1

		if self.iterator_index >= len(self):
			raise StopIteration
		return self[self.iterator_index]

	def __delitem__(self, index):
		del self.samples[index]

	def __setitem__(self, index, sample):
		self.samples[index] = sample

	def __getitem__(self, index):
		return self.samples[index]

	def get_input():
		if self.data is None or Header.INPUT_OFFSET not in self.headers or Header.INPUT_LENGTH not in self.headers:
			return None
		return self.__subdata(self.headers[Header.INPUT_OFFSET], self.headers[Header.INPUT_LENGTH])

	def get_output():
		if self.data is None or Header.OUTPUT_OFFSET not in self.headers or Header.OUTPUT_LENGTH not in self.headers:
			return None
		return self.__subdata(self.headers[Header.OUTPUT_OFFSET], self.headers[Header.OUTPUT_LENGTH])

	def get_key():
		if self.data is None or Header.KEY_OFFSET not in self.headers or Header.KEY_LENGTH not in self.headers:
			return None
		return self.__subdata(self.headers[Header.KEY_OFFSET], self.headers[Header.KEY_LENGTH])

	def __subdata(self, offset, length):
		return self.data[offset : offset + length]

	def __repr__(self):
		# Represent the data
		if len(self.data) >= 2:
			data = ', {0:02X}...{1:01X}'.format(self.data[0], self.data[-1])
		elif len(self.data) == 1:
			data = ', {0:02X}'.format(self.data[0])
		else:
			data = ''

		# Return the representation
		return '<Trace: {0:d}, {1:s}{2:s}>'.format(len(self.samples), self.title, data)



import unittest
import trsfile
import os
from pathlib import Path
from trsfile import TrsFile

class TestGeneric(unittest.TestCase):
	def test_open(self):
		with trsfile.open(Path(__file__).parent / 'data' / '90x500xfloat.trs') as trs_file:
			self.assertIsInstance(trs_file, TrsFile)

			# Assume both handles to be not closed
			self.assertFalse(trs_file.file_handle.closed)
			self.assertFalse(trs_file.handle.closed)

	def test_close(self):
		trs_file = trsfile.open(Path(__file__).parent / 'data' / '90x500xfloat.trs')
		self.assertIsInstance(trs_file, TrsFile)
		trs_file.close()

		# Assume both handles to be closed
		self.assertTrue(trs_file.file_handle.closed)
		self.assertTrue(trs_file.handle.closed)

if __name__ == '__main__':
	unittest.main()



from enum import Enum

class TracePadding(Enum):
	"""
	Defines the padding mode of the samples in each trace. This can be helpful
	when not all traces will be the same length. This can be set in
	:py:meth:`trsfile.open()`, :py:func:`trsfile.trs_open()`, :py:meth:`trsfile.trsfile_mutable.TrsFileMutable`

	+----------+---------------------------------------------------------------+
	| Mode     | Description                                                   |
	+==========+===============================================================+
	| NONE     | No padding will be used and an exception will be thrown when  |
	|          | traces are not of the same length.                            |
	+----------+---------------------------------------------------------------+
	| PAD      | All traces will be padded with zeroes to the maximum trace    |
	|          | length.                                                       |
	+----------+---------------------------------------------------------------+
	| TRUNCATE | All traces will be truncated to the minimum trace length.     |
	+----------+---------------------------------------------------------------+
	"""

	NONE = 0
	PAD = 1
	TRUNCATE = 2

class SampleCoding(Enum):
	"""
	Defines the encoding of all the samples in the trace.
	Bit 4 specifies if it is a float (1) or an integer (0),
	bits 0 to 3 specifies the length of the value.
	Finally, bits 5-7 are currently reserved and set to 000.

	This class is just a simple lookup table.
	"""

	BYTE  = (0x01, 1, 'int8')
	SHORT = (0x02, 2, 'int16')
	INT   = (0x04, 4, 'int32')
	FLOAT = (0x14, 4, 'float32')

	def __new__(cls, coding, size, format):
		obj = object.__new__(cls)
		obj._value_ = coding
		obj.size = size
		obj.format = format
		return obj

	@property
	def is_float(self):
		return (self._value_ & 0x10) != 0

class Header(Enum):
	NUMBER_TRACES                    = (0x41, 'NT', True,  int,          4,      None,    'Number of traces')
	NUMBER_SAMPLES                   = (0x42, 'NS', True,  int,          4,      None,    'Number of samples per trace')
	SAMPLE_CODING                    = (0x43, 'SC', True,  SampleCoding, 1,      None,    'Sample Coding (see SampleCoding class)')
	LENGTH_DATA                      = (0x44, 'DS', False, int,          2,      0,       'Length of cryptographic data included in trace')
	TITLE_SPACE                      = (0x45, 'TS', False, int,          1,      0,       'Title space reserved per trace')
	TRACE_TITLE                      = (0x46, 'GT', False, str,          None,   'trace', 'Global trace title')
	DESCRIPTION                      = (0x47, 'DC', False, str,          None,   None,    'Description')
	OFFSET_X                         = (0x48, 'XO', False, int,          4,      0,       'Offset in X-axis for trace representation')
	LABEL_X                          = (0x49, 'XL', False, str,          None,   None,    'Label of X-axis')
	LABEL_Y                          = (0x4A, 'YL', False, str,          None,   None,    'Label of Y-axis')
	SCALE_X                          = (0x4B, 'XS', False, float,        4,      1,       'Scale value for X-axis')
	SCALE_Y                          = (0x4C, 'YS', False, float,        4,      1,       'Scale value for Y-axis')
	TRACE_OFFSET                     = (0x4D, 'TO', False, int,          4,      0,       'Trace offset for displaying trace numbers')
	LOGARITHMIC_SCALE                = (0x4E, 'LS', False, int,          1,      0,       'Logarithmic scale')
	ACQUISITION_RANGE_OF_SCOPE       = (0x55, 'RG', False, float,        4,      0,       'Range of the scope used to perform acquisition')
	ACQUISITION_COUPLING_OF_SCOPE    = (0x56, 'CL', False, int,          4,      0,       'Coupling of the scope used to perform acquisition')
	ACQUISITION_OFFSET_OF_SCOPE      = (0x57, 'OS', False, float,        4,      0,       'Offset of the scope used to perform acquisition')
	ACQUISITION_INPUT_IMPEDANCE      = (0x58, 'II', False, float,        4,      0,       'Input impedance of the scope used to perform acquisition')
	ACQUISITION_DEVICE_ID            = (0x59, 'AI', False, bytes,        None,   None,    'Device ID of the scope used to perform acquisition')
	ACQUISITION_TYPE_FILTER          = (0x5A, 'FT', False, int,          4,      0,       'The type of filter used during acquisition')
	ACQUISITION_FREQUENCY_FILTER     = (0x5B, 'FF', False, float,        4,      0,       'Frequency of the filter used during acquisition')
	ACQUISITION_RANGE_FILTER         = (0x5C, 'FR', False, float,        4,      0,       'Range of the filter used during acquisition')
	TRACE_BLOCK                      = (0x5F, 'TB', True,  None,         0,      None,    'Trace block marker: an empty TLV that marks the end of the header')
	EXTERNAL_CLOCK_USED              = (0x60, 'EU', False, bool,         1,      False,   'External clock used')
	EXTERNAL_CLOCK_THRESHOLD         = (0x61, 'ET', False, float,        4,      0,       'External clock threshold')
	EXTERNAL_CLOCK_MULTIPLIER        = (0x62, 'EM', False, int,          4,      0,       'External clock multiplier')
	EXTERNAL_CLOCK_PHASE_SHIFT       = (0x63, 'EP', False, int,          4,      0,       'External clock phase shift')
	EXTERNAL_CLOCK_RESAMPLER_MASK    = (0x64, 'ER', False, int,          4,      0,       'External clock resampler mask')
	EXTERNAL_CLOCK_RESAMPLER_ENABLED = (0x65, 'RE', False, bool,         1,      False,   'External clock resampler enabled')
	EXTERNAL_CLOCK_FREQUENCY         = (0x66, 'EF', False, float,        4,      0,       'External clock frequency')
	EXTERNAL_CLOCK_BASE              = (0x67, 'EB', False, int,          4,      0,       'External clock time base')
	NUMBER_VIEW                      = (0x68, 'VT', False, int,          4,      0,       'View number of traces: number of traces to show on opening')
	TRACE_OVERLAP                    = (0x69, 'OV', False, bool,         1,      False,   'Overlap: whether to overlap traces in case of multi trace view')
	GO_LAST_TRACE                    = (0x6A, 'GL', False, bool,         1,      False,   'Go to last trace on opening')
	INPUT_OFFSET                     = (0x6B, 'IO', False, int,          4,      0,       'Input data offset in trace data')
	OUTPUT_OFFSET                    = (0x6C, 'OO', False, int,          4,      0,       'Output data offset in trace data')
	KEY_OFFSET                       = (0x6D, 'KO', False, int,          4,      0,       'Key data offset in trace data')
	INPUT_LENGTH                     = (0x6E, 'IL', False, int,          4,      0,       'Input data length in trace data')
	OUTPUT_LENGTH                    = (0x6F, 'OL', False, int,          4,      0,       'Output data length in trace data')
	KEY_LENGTH                       = (0x70, 'KL', False, int,          4,      0,       'Key data length in trace data')

	def __new__(cls, tag, tag_name, is_mandatory, type, length, default, description):
		obj = object.__new__(cls)
		obj._value_ = tag
		obj.tag_name = tag_name
		obj.is_mandatory = is_mandatory
		obj.type = type
		obj.length = length
		obj.default = default
		obj.description = description
		return obj

	@classmethod
	def has_value(cls, tag):
		return any(tag == item.value for item in cls)

	@classmethod
	def get_mandatory(cls):
		return set([header for header in cls if header.is_mandatory])



import unittest
import trsfile
import os
import binascii
import numpy
from pathlib import Path
from trsfile import TrsFile, Header, SampleCoding

class TestTrsFile(unittest.TestCase):
	def setUp(self):
		self.trs_file = trsfile.open(Path(__file__).parent / 'data' / '90x500xfloat.trs')

	def tearDown(self):
		self.trs_file.close()

	def test_iterator(self):
		"""Check if we can iterate over the samples"""
		i = 0
		for samples in self.trs_file[0]:
			i += 1
		self.assertEqual(i, len(self.trs_file[0]))

	def test_data(self):
		self.assertEqual(self.trs_file[0].data, binascii.unhexlify('43B94E34D3A221B27640C5AD87FBE5DF'))

	def test_title(self):
		self.assertEqual(self.trs_file[0].title, 'Clipped trace')

if __name__ == '__main__':
	unittest.main()



import setuptools
import trsfile

with open("README.md", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name                          = trsfile.__name__,
	version                       = trsfile.__version__,
	author                        = trsfile.__author__,
	author_email                  = trsfile.__contact__,
	description                   = (
		"Library to read and create Riscure Inspector trace set files (.trs)"
	),
	long_description              = long_description,
	long_description_content_type = "text/markdown",
	url                           = "https://github.com/riscure/python-trsfile",
	packages                      = setuptools.find_packages(),
	install_requires=[
		'numpy',
	],
	license                       = "BSD 3-Clause Clear License",
	keywords                      = "trs trace inspector riscure",
	classifiers                   = (
		"Development Status :: 4 - Beta",
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: BSD License",
		"Topic :: Utilities",
		"Operating System :: OS Independent",
	),
	project_urls                  = {
		'Documentation': 'https://trsfile.readthedocs.io/en/latest',
		'Bug Reports'  : 'https://github.com/riscure/python-trsfile/issues',
		'Riscure'      : 'https://www.riscure.com',
	},
)



import os
import struct
import shutil
import numpy
import math

from .trace import Trace
from .common import Header, SampleCoding, TracePadding
from .trsfile import TrsFile

class TrsFileMutable(TrsFile):
	temp_folder = None
	is_saved = False

	def __init__(self, path, padding_mode = TracePadding.PAD, force_overwrite = False):
		# Check if trs file already exists, warn if it does
		if not force_overwrite and os.path.isfile(path):
			raise IOError('trs file already exists, use force_overwrite argument to overwrite.')

		self.path = path
		self.padding_mode = padding_mode

		# Shadow list of traces in files
		self.shadow_trace_index = -1
		self.shadow_traces = []

		# Create the temporary folder
		self.file_path, self.file_ext = os.path.splitext(self.path)
		if len(self.file_ext) <= 1:
			self.file_ext = '.trs'
		self.temp_folder = self.file_path + '.tmp'

		# Make sure to always start with a fresh start
		if os.path.isdir(self.temp_folder):
			shutil.rmtree(self.temp_folder, True)
		os.mkdir(self.temp_folder)

		self.__initialize_headers()

	def __del__(self):
		if self.temp_folder is not None and os.path.isdir(self.temp_folder):
			shutil.rmtree(self.temp_folder, True)

	def __enter__(self):
		"""Called when entering a `with` block"""

		if not os.path.isdir(self.temp_folder):
			raise ValueError('temp folder for traces does not exists')
		return self

	def __len__(self):
		"""Returns the total number of traces"""

		return len(self.shadow_traces)

	def __delitem__(self, index):
		self.is_saved = False
		# Remove the shadow traces and with that check if indexes are correct
		exception = None
		try:
			indices = self.shadow_traces[index]
			del self.shadow_traces[index]

			if not isinstance(index, slice):
				indices = [indices]
		except IndexError as err:
			exception = err

		# Delete all traces on the file system
		for trace_index in indices:
			for category in ['title', 'data', 'samples']:
				path = self.__get_temp_trace(trace_index, category)
				if os.path.isfile(path):
					os.remove(path)

		# Do we have an exception, re-raise
		if exception is not None:
			raise IndexError(exception)

	def __getitem__(self, index):
		self.is_saved = False
		traces = []

		# Try access, and re-raise if wrong for fancy indexing errors
		try:
			indices = self.shadow_traces[index]
			if not isinstance(index, slice):
				indices = [indices]
		except IndexError as exception:
			raise IndexError(exception)

		# Now obtain all requested traces from file
		for i in indices:
			samples = None

			# Read the samples
			path = self.__get_temp_trace(i, 'samples')
			if os.path.isfile(path):
				with open(path, 'rb') as tmp_file:
					# First byte is always sample coding
					sample_coding = SampleCoding(tmp_file.read(1)[0])
					samples = numpy.fromfile(tmp_file, sample_coding.format, -1)

			# Title
			path = self.__get_temp_trace(i, 'title')
			if os.path.isfile(path):
				with open(path, 'rb') as tmp_file:
					title = tmp_file.read().decode('utf-8')
			else:
				title = Header.TRACE_TITLE.default

			# Read the data
			path = self.__get_temp_trace(i, 'data')
			if os.path.isfile(path):
				with open(path, 'rb') as tmp_file:
					data = tmp_file.read()
			else:
				data = b''

			# Sanity check
			if samples is None:
				raise IOError('unable to read samples from trace {0:d}'.format(i))

			traces.append(Trace(sample_coding, samples, data, title, self.headers))

		# Return the list of traces, or single trace
		if isinstance(index, slice):
			return traces
		else:
			# Earlier logic should ensure traces contains one element!
			return traces[0]

	def __setitem__(self, index, traces):
		self.is_saved = False

		# Make sure we have iterable traces
		if isinstance(traces, Trace):
			traces = [traces]

		# Store all traces with the next sequence numbers and keep these numbers as a list
		new_traces = []
		for trace in traces:
			# Make sure we do not do something weird
			if not isinstance(trace, Trace):
				raise ValueError('the data added to the trace set is not of type Trace')

			self.shadow_trace_index += 1
			new_traces.append(self.shadow_trace_index)

			# Save the trace data
			# Write the title as ascii
			with open(self.__get_temp_trace(self.shadow_trace_index, 'title'), 'wb') as tmp_file:
				tmp_file.write(trace.title if not isinstance(trace.title, str) else trace.title.encode('utf-8'))

			# Write the data file
			if trace.data is not None and len(trace.data) > 0:
				with open(self.__get_temp_trace(self.shadow_trace_index, 'data'), 'wb') as tmp_file:
					tmp_file.write(trace.data)

			# Write the sample file
			with open(self.__get_temp_trace(self.shadow_trace_index, 'samples'), 'wb') as tmp_file:
				tmp_file.write(bytes([trace.sample_coding.value]))
				numpy.array(trace.samples).astype(trace.sample_coding.format).tofile(tmp_file)

		# Lets delete the files that we are replacing to prevent storage explosions
		try:
			indices = self.shadow_traces[index]
			if not isinstance(index, slice):
				indices = [indices]
		except IndexError as exception:
			raise IndexError(exception)

		for i in indices:
			for category in ['title', 'data', 'samples']:
				path = self.__get_temp_trace(i, category)
				if os.path.isfile(path):
					os.remove(path)

		# Now we just assign the new_traces however, the slicing works
		if isinstance(index, slice):
			self.shadow_traces[index] = new_traces
		else:
			if len(new_traces) != 1:
				raise TypeError('assigning multiple new traces to single trace')
			self.shadow_traces[index] = new_traces[0]

	def save(self, padding_mode = None):
		# Do not save twice, just a waste!
		if self.is_saved:
			return

		# Get argument or default
		padding_mode = self.padding_mode if padding_mode is None else padding_mode

		# Calculate some fields
		check_values = {}
		check_values[Header.NUMBER_SAMPLES] = set([len(trace) for trace in self])
		check_values[Header.SAMPLE_CODING]  = set([trace.sample_coding for trace in self])
		check_values[Header.LENGTH_DATA]    = set([len(trace.data) for trace in self])

		# Calculate headers based on shadow list
		for header in Header.get_mandatory():
			if not header in self.headers:
				self.headers[header] = header.default

		# Check padding mode
		if padding_mode == TracePadding.PAD:
			check_values[Header.NUMBER_SAMPLES] = [max(check_values[Header.NUMBER_SAMPLES])]
		elif padding_mode == TracePadding.TRUNCATE:
			check_values[Header.NUMBER_SAMPLES] = [min(check_values[Header.NUMBER_SAMPLES])]

		# Set the headers and make sure all traces share the common header settings
		self.headers[Header.NUMBER_TRACES]  = len(self)
		self.headers[Header.TITLE_SPACE]    = max([len(trace.title) for trace in self])
		for header, check_values in check_values.items():
			if len(check_values) != 1:
				raise TypeError('traces have different values for header {0:s}'.format(header.name))
			self.headers[header] = check_values.pop()

		# Save the file
		with open(self.path, 'wb') as trs_file:

			# Save the headers
			for header, value in self.headers.items():
				# Skip TRACE_BLOCK header as we write that last!
				if header == Header.TRACE_BLOCK:
					continue

				# Check if we have definitions for all headers
				if not isinstance(header, Header):
					# TODO: Make this proper
					print('Warning, can not write unknown header X')
					continue

				# Obtain the tag value
				if header.type is int:
					tag_value = value.to_bytes(header.length, 'little')
				elif header.type is float:
					tag_value = struct.pack('<f', value)
				elif header.type is bool:
					tag_value = struct.pack('<?', value)
				elif header.type is str:
					tag_value = value.encode('utf-8')
				elif header.type is SampleCoding:
					tag_value = value.value.to_bytes(1, 'little')
				elif header.type is bytes:
					tag_value = value
				else:
					print('Warning, missing bla, to bla', header, header.type)
					continue

				# Write the Tag
				trs_file.write(bytes([header.value]))

				# Write the Length
				tag_length = len(tag_value)
				if tag_length >= 0x80:
					tag_length_length = math.ceil(x.bit_length() / 8.0)
					trs_file.write(bytes([0x80 | tag_length_length]) + tag_length.to_bytes(tag_length_length, 'little'))
				else:
					trs_file.write(bytes([tag_length]))

				# Write the value
				trs_file.write(tag_value)

			# Save the TRACE_BLOCK
			trs_file.write(bytes([Header.TRACE_BLOCK.value, 0]))

			# Start saving all traces
			for trace in self:
				# Title and title padding
				title = trace.title.encode('utf-8')
				trs_file.write(title)
				if len(title) < self.headers[Header.TITLE_SPACE]:
					trs_file.write(bytes([0] * (self.headers[Header.TITLE_SPACE] - len(title))))

				# Data
				trs_file.write(trace.data)

				# Automatic truncate
				numpy.array(trace.samples[:self.headers[Header.NUMBER_SAMPLES]]) \
				.astype(self.headers[Header.SAMPLE_CODING].format).tofile(trs_file)

				# Add any required padding
				if len(trace.samples) < self.headers[Header.NUMBER_SAMPLES]:
					numpy.array([0] * (self.headers[Header.NUMBER_SAMPLES] - len(trace.samples))) \
					.astype(self.headers[Header.SAMPLE_CODING].format).tofile(trs_file)

		# Set save flag to true
		self.is_saved = True

	def close(self):
		# Make sure to save any changes if we are creating a trace file
		self.save()

	def append(self, trace):
		self[len(self):len(self)] = trace

	def extend(self, traces):
		self[len(self):len(self)] = traces

	def insert(self, index, trace):
		self[index:index] = [trace]

	def __get_temp_trace(self, i, name):
		return self.temp_folder + '/{0:d}.{1:s}'.format(i, name)

	def __initialize_headers(self):
		"""Read all internal headers from the file"""
		self.headers = {}



"""
The defactor trace set (.trs files) library for reading
Riscure Inspector trace files.
"""

name        = "trsfile"
__version__ = '0.1.0'
__author__  = 'Kevin Valk'
__contact__ =  'valk@riscure.com'
__all__     = ['trs_open', 'trs_create', 'Trace', 'SampleCoding', 'Header', 'TracePadding']

from .trace import Trace
from .common import Header, SampleCoding, TracePadding
from .trsfile import TrsFile
from .trsfile_mutable import TrsFileMutable

def open(path):
	return trs_open(path)

def create(path, padding_mode = TracePadding.PAD, force_overwrite = False):
	return trs_create(path, padding_mode, force_overwrite)

def trs_open(path):
	return TrsFile(path)

def trs_create(path, padding_mode = TracePadding.PAD, force_overwrite = False):
	return TrsFileMutable(path, padding_mode, force_overwrite)
